@@ -81,9 +81,9 @@
     clippy::verbose_file_reads,
     clippy::zero_sized_map_values,
     future_incompatible,
-    missing_crate_level_docs,
     nonstandard_style,
-    rust_2018_idioms
+    rust_2018_idioms,
+    rustdoc::missing_crate_level_docs
 )]
 #![allow(clippy::float_cmp)]
 #![allow(clippy::manual_range_contains)]
@@ -95,6 +95,8 @@ pub mod file_storage;
 
 pub use egui; // Re-export for user convenience
 
+use std::sync::{Arc, Mutex};
+
 // ----------------------------------------------------------------------------
 
 /// Implement this trait to write apps that can be compiled both natively using the [`egui_glium`](https://github.com/emilk/egui/tree/master/egui_glium) crate,
@@ -104,8 +106,11 @@ pub trait App {
     ///
     /// Put your widgets into a [`egui::SidePanel`], [`egui::TopBottomPanel`], [`egui::CentralPanel`], [`egui::Window`] or [`egui::Area`].
     ///
-    /// To force a repaint, call either [`egui::Context::request_repaint`] or use [`Frame::repaint_signal`].
-    fn update(&mut self, ctx: &egui::CtxRef, frame: &mut Frame<'_>);
+    /// The [`egui::Context`] and [`Frame`] can be cloned and saved if you like.
+    ///
+    /// To force a repaint, call either [`egui::Context::request_repaint`] during the call to `update`,
+    /// or call [`Frame::request_repaint`] at any time (e.g. from another thread).
+    fn update(&mut self, ctx: &egui::Context, frame: &Frame);
 
     /// Called once before the first frame.
     ///
@@ -113,24 +118,7 @@ pub trait App {
     /// [`egui::Context::set_visuals`] etc.
     ///
     /// Also allows you to restore state, if there is a storage (required the "persistence" feature).
-    fn setup(
-        &mut self,
-        _ctx: &egui::CtxRef,
-        _frame: &mut Frame<'_>,
-        _storage: Option<&dyn Storage>,
-    ) {
-    }
-
-    /// If `true` a warm-up call to [`Self::update`] will be issued where
-    /// `ctx.memory().everything_is_visible()` will be set to `true`.
-    ///
-    /// This will help pre-caching all text, preventing stutter when
-    /// opening a window containing new glyphs.
-    ///
-    /// In this warm-up call, all painted shapes will be ignored.
-    fn warm_up_enabled(&self) -> bool {
-        false
-    }
+    fn setup(&mut self, _ctx: &egui::Context, _frame: &Frame, _storage: Option<&dyn Storage>) {}
 
     /// Called on shutdown, and perhaps at regular intervals. Allows you to save state.
     ///
@@ -145,7 +133,20 @@ pub trait App {
     /// where `APPNAME` is what is returned by [`Self::name()`].
     fn save(&mut self, _storage: &mut dyn Storage) {}
 
-    /// Called once on shutdown (before or after [`Self::save`])
+    /// Called before an exit that can be aborted.
+    /// By returning `false` the exit will be aborted. To continue the exit return `true`.
+    ///
+    /// A scenario where this method will be run is after pressing the close button on a native
+    /// window, which allows you to ask the user whether they want to do something before exiting.
+    /// See the example `eframe/examples/confirm_exit.rs` for practical usage.
+    ///
+    /// It will _not_ be called on the web or when the window is forcefully closed.
+    fn on_exit_event(&mut self) -> bool {
+        true
+    }
+
+    /// Called once on shutdown (before or after [`Self::save`]). If you need to abort an exit use
+    /// [`Self::on_exit_event`]
     fn on_exit(&mut self) {}
 
     // ---------
@@ -161,8 +162,13 @@ pub trait App {
     }
 
     /// The size limit of the web app canvas.
+    ///
+    /// By default the size if limited to 1024x2048.
+    ///
+    /// A larger canvas can lead to bad frame rates on some browsers on some platforms.
+    /// In particular, Firefox on Mac and Linux is really bad at handling large WebGL canvases:
+    /// <https://bugzilla.mozilla.org/show_bug.cgi?id=1010527#c0> (unfixed since 2014).
     fn max_size_points(&self) -> egui::Vec2 {
-        // Some browsers get slow with huge WebGL canvases, so we limit the size:
         egui::Vec2::new(1024.0, 2048.0)
     }
 
@@ -175,20 +181,34 @@ pub trait App {
         egui::Color32::from_rgba_unmultiplied(12, 12, 12, 180).into()
     }
 
-    /// Controls wether or not the native window position and size will be
+    /// Controls whether or not the native window position and size will be
     /// persisted (only if the "persistence" feature is enabled).
     fn persist_native_window(&self) -> bool {
         true
     }
 
-    /// Controls wether or not the egui memory (window positions etc) will be
+    /// Controls whether or not the egui memory (window positions etc) will be
     /// persisted (only if the "persistence" feature is enabled).
     fn persist_egui_memory(&self) -> bool {
         true
     }
+
+    /// If `true` a warm-up call to [`Self::update`] will be issued where
+    /// `ctx.memory().everything_is_visible()` will be set to `true`.
+    ///
+    /// This can help pre-caching resources loaded by different parts of the UI, preventing stutter later on.
+    ///
+    /// In this warm-up call, all painted shapes will be ignored.
+    ///
+    /// The default is `false`, and it is unlikely you will want to change this.
+    fn warm_up_enabled(&self) -> bool {
+        false
+    }
 }
 
-/// Options controlling the behavior of a native window
+/// Options controlling the behavior of a native window.
+///
+/// Only a single native window is currently supported.
 #[derive(Clone)]
 pub struct NativeOptions {
     /// Sets whether or not the window will always be on top of other windows.
@@ -210,9 +230,18 @@ pub struct NativeOptions {
     /// The application icon, e.g. in the Windows task bar etc.
     pub icon_data: Option<IconData>,
 
-    /// The initial size of the native window in points (logical pixels).
+    /// The initial (inner) position of the native window in points (logical pixels).
+    pub initial_window_pos: Option<egui::Pos2>,
+
+    /// The initial inner size of the native window in points (logical pixels).
     pub initial_window_size: Option<egui::Vec2>,
 
+    /// The minimum inner window size
+    pub min_window_size: Option<egui::Vec2>,
+
+    /// The maximum inner window size
+    pub max_window_size: Option<egui::Vec2>,
+
     /// Should the app window be resizable?
     pub resizable: bool,
 
@@ -230,7 +259,10 @@ impl Default for NativeOptions {
             decorated: true,
             drag_and_drop_support: false,
             icon_data: None,
+            initial_window_pos: None,
             initial_window_size: None,
+            min_window_size: None,
+            max_window_size: None,
             resizable: true,
             transparent: false,
         }
@@ -240,7 +272,7 @@ impl Default for NativeOptions {
 /// Image data for the icon.
 #[derive(Clone)]
 pub struct IconData {
-    /// RGBA pixels.
+    /// RGBA pixels, unmultiplied.
     pub rgba: Vec<u8>,
 
     /// Image width. This should be a multiple of 4.
@@ -254,67 +286,137 @@ pub struct IconData {
 ///
 /// It provides methods to inspect the surroundings (are we on the web?),
 /// allocate textures, and change settings (e.g. window size).
-pub struct Frame<'a>(backend::FrameBuilder<'a>);
+///
+/// [`Frame`] is cheap to clone and is safe to pass to other threads.
+#[derive(Clone)]
+pub struct Frame(pub Arc<Mutex<backend::FrameData>>);
+
+impl Frame {
+    /// Create a `Frame` - called by the integration.
+    #[doc(hidden)]
+    pub fn new(frame_data: backend::FrameData) -> Self {
+        Self(Arc::new(Mutex::new(frame_data)))
+    }
+
+    /// Access the underlying [`backend::FrameData`].
+    #[doc(hidden)]
+    #[inline]
+    pub fn lock(&self) -> std::sync::MutexGuard<'_, backend::FrameData> {
+        self.0.lock().unwrap()
+    }
 
-impl<'a> Frame<'a> {
     /// True if you are in a web environment.
     pub fn is_web(&self) -> bool {
-        self.info().web_info.is_some()
+        self.lock().info.web_info.is_some()
     }
 
     /// Information about the integration.
-    pub fn info(&self) -> &IntegrationInfo {
-        &self.0.info
-    }
-
-    /// A way to allocate textures.
-    pub fn tex_allocator(&mut self) -> &mut dyn TextureAllocator {
-        self.0.tex_allocator
+    pub fn info(&self) -> IntegrationInfo {
+        self.lock().info.clone()
     }
 
     /// Signal the app to stop/exit/quit the app (only works for native apps, not web apps).
     /// The framework will not quit immediately, but at the end of the this frame.
-    pub fn quit(&mut self) {
-        self.0.output.quit = true;
+    pub fn quit(&self) {
+        self.lock().output.quit = true;
     }
 
     /// Set the desired inner size of the window (in egui points).
-    pub fn set_window_size(&mut self, size: egui::Vec2) {
-        self.0.output.window_size = Some(size);
+    pub fn set_window_size(&self, size: egui::Vec2) {
+        self.lock().output.window_size = Some(size);
     }
 
     /// Set the desired title of the window.
-    pub fn set_window_title(&mut self, title: &str) {
-        self.0.output.window_title = Some(title.to_owned());
+    pub fn set_window_title(&self, title: &str) {
+        self.lock().output.window_title = Some(title.to_owned());
     }
 
     /// Set whether to show window decorations (i.e. a frame around you app).
     /// If false it will be difficult to move and resize the app.
-    pub fn set_decorations(&mut self, decorated: bool) {
-        self.0.output.decorated = Some(decorated);
+    pub fn set_decorations(&self, decorated: bool) {
+        self.lock().output.decorated = Some(decorated);
     }
 
     /// When called, the native window will follow the
     /// movement of the cursor while the primary mouse button is down.
     ///
-    /// Does not work on the web, and works badly on Mac.
-    pub fn drag_window(&mut self) {
-        self.0.output.drag_window = true;
+    /// Does not work on the web.
+    pub fn drag_window(&self) {
+        self.lock().output.drag_window = true;
     }
 
-    /// If you need to request a repaint from another thread, clone this and send it to that other thread.
-    pub fn repaint_signal(&self) -> std::sync::Arc<dyn RepaintSignal> {
-        self.0.repaint_signal.clone()
+    /// This signals the [`egui`] integration that a repaint is required.
+    ///
+    /// Call this e.g. when a background process finishes in an async context and/or background thread.
+    pub fn request_repaint(&self) {
+        self.lock().repaint_signal.request_repaint();
+    }
+
+    /// for integrations only: call once per frame
+    pub fn take_app_output(&self) -> crate::backend::AppOutput {
+        std::mem::take(&mut self.lock().output)
     }
 }
 
 /// Information about the web environment (if applicable).
 #[derive(Clone, Debug)]
 pub struct WebInfo {
-    /// e.g. "#fragment" part of "www.example.com/index.html#fragment".
+    /// Information about the URL.
+    pub location: Location,
+}
+
+/// Information about the URL.
+///
+/// Everything has been percent decoded (`%20` -> ` ` etc).
+#[derive(Clone, Debug)]
+pub struct Location {
+    /// The full URL (`location.href`) without the hash.
+    ///
+    /// Example: `"http://www.example.com:80/index.html?foo=bar"`.
+    pub url: String,
+
+    /// `location.protocol`
+    ///
+    /// Example: `"http:"`.
+    pub protocol: String,
+
+    /// `location.host`
+    ///
+    /// Example: `"example.com:80"`.
+    pub host: String,
+
+    /// `location.hostname`
+    ///
+    /// Example: `"example.com"`.
+    pub hostname: String,
+
+    /// `location.port`
+    ///
+    /// Example: `"80"`.
+    pub port: String,
+
+    /// The "#fragment" part of "www.example.com/index.html?query#fragment".
+    ///
     /// Note that the leading `#` is included in the string.
     /// Also known as "hash-link" or "anchor".
-    pub web_location_hash: String,
+    pub hash: String,
+
+    /// The "query" part of "www.example.com/index.html?query#fragment".
+    ///
+    /// Note that the leading `?` is NOT included in the string.
+    ///
+    /// Use [`Self::web_query_map]` to get the parsed version of it.
+    pub query: String,
+
+    /// The parsed "query" part of "www.example.com/index.html?query#fragment".
+    ///
+    /// "foo=42&bar%20" is parsed as `{"foo": "42",  "bar ": ""}`
+    pub query_map: std::collections::BTreeMap<String, String>,
+
+    /// `location.origin`
+    ///
+    /// Example: `"http://www.example.com:80"`.
+    pub origin: String,
 }
 
 /// Information about the integration passed to the use app each frame.
@@ -338,42 +440,18 @@ pub struct IntegrationInfo {
     pub native_pixels_per_point: Option<f32>,
 }
 
-/// How to allocate textures (images) to use in [`egui`].
-pub trait TextureAllocator {
-    /// Allocate a new user texture.
-    ///
-    /// There is no way to change a texture.
-    /// Instead allocate a new texture and free the previous one with [`Self::free`].
-    fn alloc_srgba_premultiplied(
-        &mut self,
-        size: (usize, usize),
-        srgba_pixels: &[egui::Color32],
-    ) -> egui::TextureId;
-
-    /// Free the given texture.
-    fn free(&mut self, id: egui::TextureId);
-}
-
 /// Abstraction for platform dependent texture reference
 pub trait NativeTexture {
     /// The native texture type.
     type Texture;
 
-    /// Bind native texture to egui texture
+    /// Bind native texture to an egui texture id.
     fn register_native_texture(&mut self, native: Self::Texture) -> egui::TextureId;
 
-    /// Change id's actual pointing texture
-    /// only for user texture
+    /// Change what texture the given id refers to.
     fn replace_native_texture(&mut self, id: egui::TextureId, replacing: Self::Texture);
 }
 
-/// How to signal the [`egui`] integration that a repaint is required.
-pub trait RepaintSignal: Send + Sync {
-    /// This signals the [`egui`] integration that a repaint is required.
-    /// This is meant to be called when a background process finishes in an async context and/or background thread.
-    fn request_repaint(&self);
-}
-
 // ----------------------------------------------------------------------------
 
 /// A place where you can store custom data in a way that persists when you restart the app.
@@ -413,10 +491,7 @@ pub fn get_value<T: serde::de::DeserializeOwned>(storage: &dyn Storage, key: &st
 /// Serialize the given value as [RON](https://github.com/ron-rs/ron) and store with the given key.
 #[cfg(feature = "ron")]
 pub fn set_value<T: serde::Serialize>(storage: &mut dyn Storage, key: &str, value: &T) {
-    storage.set_string(
-        key,
-        ron::ser::to_string_pretty(value, Default::default()).unwrap(),
-    );
+    storage.set_string(key, ron::ser::to_string(value).unwrap());
 }
 
 /// [`Storage`] key used for app
@@ -428,27 +503,29 @@ pub const APP_KEY: &str = "app";
 pub mod backend {
     use super::*;
 
+    /// How to signal the [`egui`] integration that a repaint is required.
+    pub trait RepaintSignal: Send + Sync {
+        /// This signals the [`egui`] integration that a repaint is required.
+        ///
+        /// Call this e.g. when a background process finishes in an async context and/or background thread.
+        fn request_repaint(&self);
+    }
+
     /// The data required by [`Frame`] each frame.
-    pub struct FrameBuilder<'a> {
+    pub struct FrameData {
         /// Information about the integration.
         pub info: IntegrationInfo,
-        /// A way to allocate textures (on integrations that support it).
-        pub tex_allocator: &'a mut dyn TextureAllocator,
+
         /// Where the app can issue commands back to the integration.
-        pub output: &'a mut AppOutput,
+        pub output: AppOutput,
+
         /// If you need to request a repaint from another thread, clone this and send it to that other thread.
         pub repaint_signal: std::sync::Arc<dyn RepaintSignal>,
     }
 
-    impl<'a> FrameBuilder<'a> {
-        /// Wrap us in a [`Frame`] to send to [`App::update`].
-        pub fn build(self) -> Frame<'a> {
-            Frame(self)
-        }
-    }
-
     /// Action that can be taken by the user app.
-    #[derive(Clone, Debug, Default, PartialEq)]
+    #[derive(Default)]
+    #[must_use]
     pub struct AppOutput {
         /// Set to `true` to stop the app.
         /// This does nothing for web apps.
@@ -460,10 +537,10 @@ pub mod backend {
         /// Set to some string to rename the outer window (e.g. glium window) to this title.
         pub window_title: Option<String>,
 
-        /// Set to some bool to change window decorations
+        /// Set to some bool to change window decorations.
         pub decorated: Option<bool>,
 
-        /// Set to true to drap window
+        /// Set to true to drag window while primary mouse button is down.
         pub drag_window: bool,
     }
 }
